# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
The vis package implements various visualization tools. For example, a VTK
Structure viewer.
"""
